﻿/*
Date: Dec 7, 2015
File: BlazeWing.cs
Description: Mimics Blazing Heart's blaze wing skill from Elsword. This follows only hostile peds and "hits" them once as well as burns them, then it goes to the next ped
in the radius.
Author: flyingratpigeon@gmail.com / flyingratmodding.blogspot.com
Info: feel free to contact me in case of any bugs or inquiries. 
Ver: 1.1
Planned: More blazing heart skills
*/

using System;
using GTA;
using System.Windows.Forms;
using System.Drawing;

namespace BlazeWing
{
   
    public class Wing : Script
    {
        //check what game episode the user is playing
        private GameEpisode episode = Game.CurrentEpisode;
        private const GameEpisode EP_TLAD = GameEpisode.TLAD;
        private const GameEpisode EP_GTAIV = GameEpisode.GTAIV;

        //following variables are for the ini files
        Keys key;
        int maxPeds;
        float radius;
        public Wing()
        {            
            //set keys to ini file
            key = Settings.GetValueKey("KEYPRESS", "SETTINGS", Keys.L);
            //get number of peds to stop at
            maxPeds = Settings.GetValueInteger("MAXPEDS", "SETTINGS",3);
            //get radius from ini file
            radius = Settings.GetValueInteger("RADIUS", "SETTINGS", 50);
            //bind keydown event
            KeyDown += new GTA.KeyEventHandler(pKeyDown);
        }

        //calculates how much x changes
        private double circleX(double angle)
        {
            double xChange = Math.Cos(angle);
            xChange *= .5F;
            return xChange;
        }

        //calculates how much y changes
        private double circleY(double angle)
        {
            double yChange = Math.Sin(angle);
            yChange *= .5F;
            return yChange;
        }

        //checks if the ped is any of the given types
        //if they are, even if they are hostile (which is not often) do not attack
        private bool pedThreat(RelationshipGroup pedGroup)
        {
            switch (pedGroup)
            {	
				//all these groups are non threatening (to the player) types.
				//no one here will ever be targeted even if hostile
                case RelationshipGroup.Bum:
                case RelationshipGroup.Civillian_Female:
                case RelationshipGroup.Civillian_Male:
                case RelationshipGroup.Fireman:
                case RelationshipGroup.Medic:
                case RelationshipGroup.Prostitute:
                case RelationshipGroup.Player:
                case RelationshipGroup.Special:
                    return false;
                default:
                    return true;
            }
        }

        //if the following ped models are detected, do not attack. This (should) work for even modded ped models but let me know if it doesn't
        //read readme for how to change this list to your liking. Let me know if there is a better way to do this, as I am not the most skilled
        //of coders
        private bool pedModelThreat(Ped ped)
        {
            bool threat = true;
            //checks hash for the model
            uint hashCode = (uint) ped.Model.Hash;

            //gta iv switch
            if (episode == EP_GTAIV)
            {
                //exception for Hostile Mission negotiation, this ignores the 
                //two-three models in the top room of the warehouse. This is to prevent a bug
                if (ped.CurrentRoom.ToString().Contains("R_A8F0856F_"))
                {
                    return false;
                }
                switch (hashCode)
                {
                    //jacob hashes
                    case 1487004273:
                    case 3019935309:
                    //packie hashes
                    case 1690783035:
                    //gta iv johnny
                    case 3383459612:
                    //gta iv manny
                    case 1445589009:
                    //gta iv dwayne
                    case 3677703193:
                    //roman
                    case 2302238665:
                    case 558221221:
                    //michael
                    case 735211577:
                    //derrick
                    case 1169442297:
                    //mallorie - might not be needed, I don't think she's ever in a dangerous mission
                    case 3254679890:
                    //kate
                    case 3521216458:
                    //playboy
                    case 1794146792:
                    //badman 
                    case 1495769888:
                    //bernie Crane
                    case 1500493064:
                        threat = false;
                        break;
                }
            }//end of GTAIV if
            //TLAD switch
            else if (episode == EP_TLAD)
            {
                switch (hashCode)
                {
                    //beginning of TLAD hashes
                    //niko TLAD
                    case 1613899343:
                    //clay hashes
                    case 1825562762:
                    //terry hashes
                    case 1728056212:
                    //jim hashes
                    case 870892404:
                    //malc hashes
                    case 4055673113:
                    //marta hash
                    case 2687923072:
                    //desean hashes
                    case 4221176784:
                    //billy TLAD not prison
                    case 3843248439:
                    //brian
                    case 349841464:
                    //jason TLAD
                    //first case is actually under the regular archive img rather than TLAD but the game uses the regular archive img one
                    case 170756246:
                    case 3346030785:
                    //ashley
                    case 3567004438:
                    //beginning of unimportant gang members that go to gang wars
                    case 1914397972:
                    case 2156528113:
                    case 1215631816:
                    case 1706970202:
                    case 717510247:
                    case 965080042:
                    case 693982133:
                    case 454735664:
                    case 1409362172:
                    case 767450539: 
                    case 1686719296:
                    case 1917871822:
                    case 422524045: 
                        threat = false;
                        break;
                }
            }//end of TLAD if
            else //TBOGT switch - I have not played TBOGT yet, so let me know if Luis has other noteworthy allies or people he needs to protect
            {
                switch (hashCode)
                {
                    //gracie
                    case 2014087898:
                    case 3937314702:
                    //armando
                    case 1370299619:
                    //tony
                    case 4020398429:
                    //yusuf
                    case 3846796161:
                    //henrique
                    case 1905515841:
                    //daisy
                    case 653404222:
                        threat = false;
                        break;
                }
            } //end of Game Episode ifs and TBOGT if
            return threat;
        }

        //checks if ped is in combat
        private bool combatStatus(Ped ped)
        {
            return ped.isInCombat || ped.isShooting || ped.isRequiredForMission;
        }

        //get nearest ped
        private Ped getPed(Vector3 currPos)
        {
            //set variables
            Ped[] pedList;
            double dist = 0;
            double currDist = 0;
            //by default the closest ped will be the player character
            Ped closestPed = Player.Character;

            pedList = World.GetPeds(currPos, radius);

            foreach (Ped indivPed in pedList)
            {
                //note that you can get the closest ped with the method World.getClosestPed
                //but that does not distinguish between pedestrians
                //will only target peds in combat that are not of the relationship groups in the notThreat method
                if (Game.Exists(indivPed) && !indivPed.isDead && combatStatus(indivPed) && pedThreat(indivPed.RelationshipGroup) && pedModelThreat(indivPed))
                {
                    currDist = indivPed.Position.DistanceTo(currPos);
                    //check if the peds current distance is less than the one before it
                    //if so we want to set the current closest ped to that one
                    //if this is the first ped targeted, set dist to a value other than 0
                    if (dist == 0 || currDist < dist)
                    {
                        dist = currDist;
                        closestPed = indivPed;
                    }
                }

            }
            return closestPed;
        }

        //loop path toward ped, takes particle id as a argument, ped and current position
        //lerp function can do the linear part but I am not sure how to use it
        private void flameMovement(int partID, Ped ped, Vector3 currPos)
        {
            const int STEPS = 25;//more steps = smoother movement
            double scale = ped.Position.DistanceTo(currPos)/STEPS;//this is the scale for the vector calculations.

            //create a variable to store the current position's coordinates
			//these are the variables that will increment in order to simulate movement
            double newPosX = currPos.X;
            double newPosY = currPos.Y;
            double newPosZ = currPos.Z;
			
			//this calculates the vector which is done by (x2-x1, y2-y1, z2-z1) 
            //for example if the first point is (1,1,1) and the second is (3,3,3)
            //it would be the second subtracting the first so the vector is (2,2,2)
            double dx = ped.Position.X - newPosX;
            double dy = ped.Position.Y - newPosY;
            double dz = ped.Position.Z - newPosZ;
			
			//find length of vector
            double length = Math.Sqrt((dx * dx) + (dy * dy) + (dz * dz));
			
			//the next point is determined by currPos + an offset of each coordinate
            //the offset is a normalized vector.
            //normalized vectors are calculated by dividing each coordinate by the vector length
            double xOffset = scale*(dx / length);
            double yOffset = scale*(dy / length);
            double zOffset = scale*(dz / length);

            //the below variables are for the circling effect
            const double ANGLE_RADIANS = (15 * Math.PI) / 180;
            double angle = 0;
            double circXoffset;
            double circYoffset;
            double slopeAtPoint;
            //end variables for circle

			//first for loop is the linear movement of the fireball
            for (int i = 0; i < STEPS; i++)
            {
                //increment the newPositions by the offsets each time to get new points and to create the illusion of movement
                newPosX = newPosX + xOffset;
                newPosY = newPosY + yOffset;
                newPosZ = newPosZ + zOffset;

                //this is what changes the position of the particle effect
                GTA.Native.Function.Call("UPDATE_PTFX_OFFSETS", partID, newPosX, newPosY, newPosZ, 0, -90.0, 0);

                if (Game.Exists(ped) && i > 23)//consider experimenting with ped.isOnScreen
                {
                    //if the ped doesn't get set on fire near end of loop, force it
                    ped.isOnFire = true;
                }

                Wait(25);//slow down movement just a bit

			}//end of linear movement, by now the fireball should be near or at the position of the ped

			//circle the enemy using the power of the unit circle
			for (int i = 0; i <= 50; i += 1)
			{
				angle += ANGLE_RADIANS; //increase our angle along the circle
                //these two methods calculate the position on the circle.
				circXoffset = circleX(angle);
				circYoffset = circleY(angle);
				//The slope at a point on a circle is the implicit derivation of x^2 + y^2 = radius. Taking the derivative becomes 2x + 2y*y' = 0.
				//solve for y' = -x/y, and that's where we get our formula.
				slopeAtPoint = (-circXoffset) / circYoffset;

				//the last 3 arguments are yaw, pitch and roll, in that order.
				//since yaw is rotating around the z axis, that's what we want to modify for the tail
				//peds have tendency to disappear if too far so make a check to see if it still exists
				if (Game.Exists(ped))
				{
                    ped.Health = ped.Health - 8;
					//Since we are calling the ped coordinates directly, this will follow the ped even if they move while circling them.
					GTA.Native.Function.Call("UPDATE_PTFX_OFFSETS", partID, ped.Position.X + circXoffset, ped.Position.Y + circYoffset, ped.Position.Z, slopeAtPoint, -90.0, 0);
                    Wait(25);//slow down the movement just a bit
                }
				 
			}
            if(Game.Exists(ped) && ped.isOnFire)
            {
                //put out fire because the ptfx cannot spawn so many times.
                ped.isOnFire = false;
            }
            
        }//end of flameMovement Method

        //make player temp invincible
        private void invincible()
        {
            //make player invincible for short time
            Player.Character.Invincible = true;
            Player.Character.PreventRagdoll = true;
            
        }//end of invincible function

        /*//this tests an offset to see if we're there yet.
        old function, no longer used, kept for other possible purposes
        private double testOffset(double coord, double pedCoord, double offset)
        {
            double newCoord = coord + offset;
            const double RANGE = .5;
            //this makes coord check a range so its in the general ped's area,
            //doesn't need to be exactly on the ped coords
            if (newCoord >= pedCoord - RANGE && newCoord <= pedCoord + RANGE)
            {
                //set this particular offset to 0 so that the fireball doesn't move on that dimension anymore
                //Ex: x coord of fireball is very close to x coord of ped so x no longer has an offset
                offset = 0;
            }
            return offset;
        }*/

        //this function plays an animation. Since blaze wing is activated using the left hand,
        //we can just leave the weapon on the other hand. However, we have to check if player is in
        //vehicle.
        private void playAnim()
        {
            if (!Player.Character.isInVehicle())
            {
                //create the animation set object
                AnimationSet animSet = new AnimationSet("BHSkills");
                String anim = "blaze_wing";

                //call our animation
                Player.Character.Animation.Play(animSet, anim, 1.0F);
                Wait(966); //966 ms because that is how long it takes for the character
                //to get to the outstretched arm part of the anim. We don't want the flame
                //effect to start until that part of the anim.
            }
        }
        //consider a check for assistant drivers to enemies?
        public void pKeyDown(object sender, GTA.KeyEventArgs e)
        {
            if (e.Key == key)
            {
                //declare a variable for a ped
                Ped ped;

                Game.DisplayText("Blaze Wing Activated.");

                invincible();

                playAnim();

                //get player position, this will change after one ped is dead
                Vector3 currPos = Player.Character.GetBonePosition(Bone.LeftHand);

                //make a fireball effect in front of character
                int fireBlastID = GTA.Native.Function.Call<int>("START_PTFX", "ambient_fire_generic", currPos.X, currPos.Y, currPos.Z, 0, -90.0, 0, 5.0);

                //main while loop
                
                for (int i = 0; i < maxPeds; i++)
                {
                    ped = getPed(currPos);
                    
                    //if the ped returned is the player character (default), then we break loop and do nothing.
                    if (ped == Player.Character)
                    {
                        i = maxPeds + 1; //break loop
                        Game.DisplayText("No hostile targets in area.");
                    }
                    else if(Game.Exists(ped))
                    {
						//start the flame's movement
                        flameMovement(fireBlastID, ped, currPos);
                    }
					//the reason why this is not put under flamemovement in the same if statement is because of the length of time that 
					//function runs so it is best to do another fresh check on the ped especially since this script allows for use in vehicle.
					if (Game.Exists(ped))
					{
                        //update the current position to the dead ped's position.
                        currPos = ped.Position;
					}
                }
                Wait(1000);
                GTA.Native.Function.Call("STOP_PTFX", fireBlastID);
                Wait(2000);

                //stop character invincibility
                Player.Character.Invincible = false;
                Player.Character.PreventRagdoll = false;

                Game.DisplayText("Blaze Wing has ended.");

            }//end of key if statement
        }//end of pKeyDown method
    }//end of wing class
}//end of namespace

